"""Dijkstra-Algorithmus
Vorlage zum Buch Informatik 5, Cornelsen Verlag
Kapitel 1.9, Aufgabe 3
Autor: Klaus Reinold
"""

class Knoten:
    """
    Beschreibt einen Knoten
    """
    def __init__(self, bezeichner):
        """
        Besetzt die Attribute und legt das Knotensymbol an.
        """
        self.bezeichner = bezeichner
        self.länge = 0
        self.vorgänger = -1
    
    def BezeichnerGeben(self):
        """
        Meldet den Bezeichner des Knotens zurück.
        """
        return self.bezeichner
    
    def LängeSetzen(self, l):
        """
        Setzt die Weglänge
        """
        self.länge = l
    
    def LängeGeben(self):
        """
        Meldet die Weglänge zurück.
        """
        return self.länge
    
    def VorgängerSetzen(self, v):
        """
        Setzt den Vorgänger
        """
        self.vorgänger = v
    
    def VorgängerGeben(self):
        """
        Meldet den Vorgänger zurück.
        """
        return self.vorgänger

class GraphMatrix:
    """
    Verwaltet einen ungerichteten, gewichteten Graphen mittels Adjazenzmatrix
    """
    def __init__(self):
        """
        Baut die Datenstruktur auf
        """
        self.knoten = []
        self.matrix = []
        self.warteliste = []
        self.fertigeKnoten = []

    def KnotenEinfügen(self, bezeichner):
        """
        Einfügen eines neuen Knoten in den Graphen
        """
        self.knoten.append(Knoten(bezeichner))
        for row in self.matrix:
            row.append(-1)
        self.matrix.append([-1] * len(self.knoten))

    def KnotenGeben(self, bezeichner):
        """
        Gibt den Knoten mit dem angegebenen Bezeichner zurück.
        """
        for k in self.knoten:
            if k.BezeichnerGeben() == bezeichner:
                return k
        return None

    def KnotenNummerGeben(self, bezeichner):
        """
        Gibt die interne Nummer des Knotens
        """
        for i, k in enumerate(self.knoten):
            if k.BezeichnerGeben() == bezeichner:
                return i
        return -1

    def KnotenBezeichnerGeben(self, knotenNummer):
        """
        Gibt die Bezeichnung eines Knotens mit der internen Knotennummer zurück.
        """
        if knotenNummer < len(self.knoten) and knotenNummer >= 0:
            return self.knoten[knotenNummer].BezeichnerGeben()
        else:
            return ""

    def KanteEinfügen(self, startKnoten, endKnoten, gewicht):
        """
        Einfügen einer Kante in den Graphen
        """
        startNummer = self.KnotenNummerGeben(startKnoten)
        endNummer = self.KnotenNummerGeben(endKnoten)
        self.matrix[startNummer][endNummer] = gewicht
        self.matrix[endNummer][startNummer] = gewicht
        
    def KnotenAnzahlgeben(self):
        """
        Gibt die Anzahl der Knoten zurück
        """
        return len(self.knoten)
    
    def KanteGewichtGeben(self, von, nach):
        """
        Gibt das Gewicht einer Kante zurück
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        
        if vonNummer != -1 and nachNummer != -1:
            return self.matrix[vonNummer][nachNummer]
        else:
            return -1
    
    def ZurückSetzen(self):
        """
        Setzt alles zurück
        """
        self.knoten.clear()
        self.matrix.clear()
        
    def DijkstraAusführen(self, startKnoten, zielKnoten):
        """
        Der Dijkstra-Algorithmus
        """
        # Vorbelegung
        warteliste = []
        fertigeKnoten = []
        self.knoten[startKnoten].LängeSetzen(0)
        self.knoten[startKnoten].VorgängerSetzen(-1)
        
        # es geht los mit dem Startknoten
        aktuellerKnoten = startKnoten
        
        # solange Ziel nicht erreicht...
        while aktuellerKnoten != zielKnoten and aktuellerKnoten != -1:
            # untersuche alle Knoten
            for nummer in range(len(self.knoten)):
                # suche den Eintrag in der Adjazenzmatrix mit den Koordinaten (aktuellerKnoten|nummer)
                längeNeu = self.matrix[aktuellerKnoten][nummer]
                # falls es eine Kante gibt und der Knoten noch nicht fertig ist...
                if längeNeu > 0 and nummer not in fertigeKnoten:
                    # überprüfe, ob der Knoten bereits einen Vorgänger hat
                    if nummer in warteliste:
                        # überprüfe, ob es über den aktuellen Knoten eine kürzere Verbindung gibt als die bisher gefundene
                        if self.knoten[aktuellerKnoten].LängeGeben() + längeNeu < self.knoten[nummer].LängeGeben():
                            # falls ja, aktualisiere die Länge und den Vorgänger des Knoten
                            self.knoten[nummer].LängeSetzen(self.knoten[aktuellerKnoten].LängeGeben() + längeNeu)
                            self.knoten[nummer].VorgängerSetzen(aktuellerKnoten)
                    else:
                        # falls der Knoten bisher keinen Vorgänger hat, wird er in die Warteliste aufgenommen.
                        warteliste.append(nummer)
                        self.knoten[nummer].VorgängerSetzen(aktuellerKnoten)
                        self.knoten[nummer].LängeSetzen(self.knoten[aktuellerKnoten].LängeGeben() + längeNeu)
            
            # der Knoten ist fertig
            fertigeKnoten.append(aktuellerKnoten)
            
            # Ausgabe des Zwischenstandes als Tabelle wie im Schulbuch
            print("Ort".ljust(20), "kommt von".ljust(20), "Distanz zum Start".ljust(20), "bearbeitet".ljust(20))
            for k in self.knoten:
                print(k.BezeichnerGeben().ljust(20), end="")
                if k.VorgängerGeben() != -1:
                    print(self.knoten[k.VorgängerGeben()].BezeichnerGeben().ljust(20), end="")
                else:
                    print("kein Vorgänger".ljust(20), end="")
                print(str(k.LängeGeben()).ljust(20), end="")
                if self.KnotenNummerGeben(k.BezeichnerGeben()) in fertigeKnoten:
                    print("ok".ljust(20))
                else:
                    print()
            print()
            print()
            
            # nun wird der nächste Knoten gesucht: ein bisher nicht besuchter Knoten mit minimaler Distanz zum Startknoten
            indexOptimalerKnoten = 0
            for nummer in range(1, len(warteliste)):
                if self.knoten[warteliste[indexOptimalerKnoten]].LängeGeben() > self.knoten[warteliste[nummer]].LängeGeben():
                    indexOptimalerKnoten = nummer
            # falls noch Knoten vorhanden sind, wird der gerade gesuchte Knoten entfernt aus der Warteliste und als 
            # aktueller Knoten für die nächste Runde festgelegt. 
            if len(warteliste) > 0:
                aktuellerKnoten = warteliste[indexOptimalerKnoten]
                warteliste.pop(indexOptimalerKnoten)
            else:
                aktuellerKnoten = -1
        # Ausgabe des kürzesten Weges
        if aktuellerKnoten != -1:
            print("Der Weg hat die Länge", self.knoten[aktuellerKnoten].LängeGeben())
            namen = []
            while aktuellerKnoten != -1:
                namen.append(self.knoten[aktuellerKnoten].BezeichnerGeben())
                aktuellerKnoten = self.knoten[aktuellerKnoten].VorgängerGeben()
            for index in range(len(namen)-1, -1, -1):
                print(namen[index])
        else:
            print("Ziel nicht erreichbar")
            

def Dijkstra_Test():
    """
    Test des Dijkstra-Algorithmus am Beispiel aus dem Schulbuch
    """
    g = GraphMatrix()
    g.KnotenEinfügen("Traunstein")
    g.KnotenEinfügen("Innsbruck")
    g.KnotenEinfügen("Innichen")
    g.KnotenEinfügen("Millstatt")
    g.KnotenEinfügen("Venedig")
    g.KanteEinfügen("Traunstein", "Innsbruck", 138)
    g.KanteEinfügen("Traunstein", "Millstatt", 211)
    g.KanteEinfügen("Innsbruck", "Innichen", 139)
    g.KanteEinfügen("Innichen", "Millstatt", 120)
    g.KanteEinfügen("Innichen", "Venedig", 197)
    g.KanteEinfügen("Millstatt", "Venedig", 278)
    g.DijkstraAusführen(g.KnotenNummerGeben("Traunstein"), g.KnotenNummerGeben("Venedig"))
        
Dijkstra_Test()

        